/*
 * Copyright 2023-2025 Hartmut Seichter and Contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

// based on image_loader by Robin Rottstädt

#include "paradiso/bitmap_io.hpp"
#include "paradiso/bitmap.hpp"

#include <filesystem>
#include <string>
#include <unordered_map>

// STB image loading
#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

namespace paradiso {

struct BitmapIO::Impl {

    std::filesystem::path asset_path_ = {};

    using BitmapCacheType = std::unordered_map<std::string_view, Bitmap>;

    BitmapCacheType cache_;

    static BitmapIO::LoadResult _read_bitmap(const path& filepath) {

        // Load with stb_image
        stbi_set_flip_vertically_on_load(true); // flip y axis for OpenGL

        int width{}, height{}, channels{};

        auto stb_free = [](uint8_t* data) { stbi_image_free(data); };

        if (std::unique_ptr<uint8_t, decltype(stb_free)> image{stbi_load(
                filepath.string().c_str(), &width, &height, &channels, 0)};
            image.get()) {

            const auto area = width * height;

            // Convert to Vector of RGBA
            std::vector<paradiso::RGBA> rgba =
                std::vector<paradiso::RGBA>(area);
            for (int i = 0; i < area; i++) {
                // get rgba values
                int pos = i * 4;
                int r = image.get()[pos + 0];
                int g = image.get()[pos + 1];
                int b = image.get()[pos + 2];
                int a = image.get()[pos + 3];

                // bug in from_rgba. it expects bgra, not rgba
                // auto val = paradiso::RGBA::from_rgba(b, g, r, a);
                rgba[i] = paradiso::RGBA::from_rgba(r, g, b, a);
            }
            return paradiso::Bitmap::from_data(
                paradiso::Size{static_cast<unsigned int>(height),
                               static_cast<unsigned int>(width)},
                rgba);

            // rotate image by 90 degrees CW
            // std::vector<paradiso::RGBA> rotated =
            //     std::vector<paradiso::RGBA>(size);
            // for (int x = 0; x < width; x++) {
            //     for (int y = 0; y < height; y++) {

            //         int pos = y * width + x;
            //         int pos_rotated = x * height + (height - y - 1);
            //         rotated[pos_rotated] = rgba[pos];
            //     }
            // }

            // return paradiso::Bitmap::from_data(
            //     paradiso::Size{static_cast<unsigned int>(height),
            //                    static_cast<unsigned int>(width)},
            //     rotated);
        } else {
            return std::unexpected(BitmapIO::LoadError::FileNotFound);
        }

        return Bitmap(); //  well, no exception ...
    }

    auto load(std::string_view filename, bool ignore_cache) -> LoadResult {
        // check cache first
        if (cache_.find(filename) != cache_.end() && !ignore_cache) {
            return cache_[filename.data()];
        } // construct file path
        auto asset_filepath = asset_path_ / std::filesystem::path(filename);

        auto data = _read_bitmap(asset_filepath);

        if (data.has_value()) {
            cache_[filename.data()] = *data;
        }
        return data;
    }
};

BitmapIO::BitmapIO() : impl_{std::make_unique<BitmapIO::Impl>()} {}

BitmapIO& BitmapIO::get() {
    static BitmapIO instance;
    return instance;
}

auto BitmapIO::load(std::string_view filename, bool ignore_cache) const
    -> BitmapIO::LoadResult {
    return impl_->load(filename, ignore_cache);
}

void BitmapIO::set_asset_path(BitmapIO::path path_) {
    impl_->asset_path_ = path_;
}

BitmapIO::path BitmapIO::asset_path() const { return impl_->asset_path_; }

} // namespace paradiso
